/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;

import org.ow2.orchestra.designer.bpmn.model.AbstractElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.SequenceFlowModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;

/**
 * Transforms {@link SequenceFlowModel}s into {@link TSequenceFlow}s and vice
 * versa
 *
 * @author Loic Albertin
 */
public final class SequenceFlowTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private SequenceFlowTransformer() {

  }

  /**
   * @param tSequenceFlow
   * @param context
   * @return
   * @throws TransformationException
   */
  public static void bpmnToModel(final TSequenceFlow tSequenceFlow, final Bpmn2DesignerContext context)
  throws TransformationException {
    final SequenceFlowModel sequenceFlowModel = new SequenceFlowModel();
    sequenceFlowModel.setProcessModel(context.getProcessModel());
    sequenceFlowModel.setName(tSequenceFlow.getName());
    sequenceFlowModel.setId(tSequenceFlow.getId());

    // Checks source and target
    if (!(tSequenceFlow.getSourceRef() instanceof TFlowElement)) {
      throw new TransformationException("BPMN conversion error: sourceRef attribute of sequenceFlow " + tSequenceFlow.getId()
          + " is not a FlowElement.");
    }

    if (!(tSequenceFlow.getTargetRef() instanceof TFlowElement)) {
      throw new TransformationException("BPMN conversion error: targetRef attribute of sequenceFlow " + tSequenceFlow.getId()
          + " is not a FlowElement.");
    }
    // Add to list before calling ElementTransformer.bpmnToModel (can cause recursion and duplicate elements)
    context.getModelsMap().put(tSequenceFlow, sequenceFlowModel);

    final TFlowElement sourceElement = (TFlowElement) tSequenceFlow.getSourceRef();
    ElementTransformer.bpmnToModel(sourceElement, context);

    final TFlowElement targetElement = (TFlowElement) tSequenceFlow.getTargetRef();
    ElementTransformer.bpmnToModel(targetElement, context);

    if (!(context.getModelsMap().get(tSequenceFlow.getSourceRef()) instanceof AbstractElementWithPosition)) {
      throw new TransformationException("BPMN conversion error: sourceRef attribute of sequenceFlow " + tSequenceFlow.getId()
          + " is not a valid FlowElement.");
    }

    if (!(context.getModelsMap().get(tSequenceFlow.getTargetRef()) instanceof AbstractElementWithPosition)) {
      throw new TransformationException("BPMN conversion error: targetRef attribute of sequenceFlow " + tSequenceFlow.getId()
          + " is not a valid FlowElement.");
    }
    sequenceFlowModel.setSourceElement((AbstractElementWithPosition) context.getModelsMap().get(tSequenceFlow.getSourceRef()));
    sequenceFlowModel.setTargetElement((AbstractElementWithPosition) context.getModelsMap().get(tSequenceFlow.getTargetRef()));

    context.getProcessModel().getElements().add(sequenceFlowModel);

    // Parse condition expression
    if (tSequenceFlow.getConditionExpression() != null
        && !tSequenceFlow.getConditionExpression().getContent().isEmpty()) {
      sequenceFlowModel.setCondition(tSequenceFlow.getConditionExpression().getContent().get(0).toString());
    }

    // Parse isDefault
    if (sourceElement instanceof TExclusiveGateway) {
      final TExclusiveGateway sourceExclusiveGateway = (TExclusiveGateway) sourceElement;
      if (tSequenceFlow.equals(sourceExclusiveGateway.getDefault())) {
        sequenceFlowModel.setIsDefault(true);
      }
      if (sequenceFlowModel.getCondition() == null) {
        sequenceFlowModel.setCondition("");
      }
    }
    // Currently we don't care about the diagram part of the sequence flow
  }

}
