/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.TaskModel;
import org.ow2.orchestra.designer.bpmn.model.TaskType;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.Assignment;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.DataOutput;
import org.ow2.orchestra.jaxb.bpmn.DataOutputAssociation;
import org.ow2.orchestra.jaxb.bpmn.InputSet;
import org.ow2.orchestra.jaxb.bpmn.IoSpecification;
import org.ow2.orchestra.jaxb.bpmn.MessageFlow;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.OutputSet;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.TCollaboration;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TExpression;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TFormalExpression;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;
import org.ow2.orchestra.jaxb.bpmn.TServiceTask;
import org.ow2.orchestra.jaxb.bpmn.TTask;

/**
 * Transform {@link TaskModel}s into {@link TTask}s
 *
 * @author Guillaume Porcher
 */
public final class TaskTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private TaskTransformer() {

  }


  public static void modelToBpmn(final TaskModel taskModel, final Designer2BpmnContext context) throws TransformationException {
    if (taskModel.getId() == null || taskModel.getId().equals("")) {
      throw new TransformationException("BPMN conversion error: A task has no id.");
    }
    final TTask tTask;
    final JAXBElement<? extends TTask> jaxbElement;

    if (taskModel.getTaskType().equals(TaskType.SERVICE)) {
      tTask = new TServiceTask();
      jaxbElement = context.getBpmnObjectFactory().createServiceTask((TServiceTask) tTask);

      // Add or get operation
      if (taskModel.getOperation() != null) {
        TMessage inMessage = null;
        TMessage outMessage = null;
        final Operation operation = TaskTransformer.modelToBpmnOperation(taskModel, context);
        ((TServiceTask) tTask).setOperationRef(new QName(context.getDefinitions().getTargetNamespace(), operation.getId()));
        // get associated messages
        for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
          if (rootElement.getValue() instanceof TMessage) {
            if (((TMessage) rootElement.getValue()).getId().equals(operation.getInMessageRef().getLocalPart())) {
              inMessage = (TMessage) rootElement.getValue();
            }
            if (operation.getOutMessageRef() != null &&
                ((TMessage) rootElement.getValue()).getId().equals(operation.getOutMessageRef().getLocalPart())) {
              outMessage = (TMessage) rootElement.getValue();
            }
          }
        }

        // Add message flows
        QName taskQName = new QName(context.getDefinitions().getTargetNamespace(), taskModel.getId());
        QName poolQName =
                new QName(context.getDefinitions().getTargetNamespace(),
                        taskModel.getOperation().getInterfaceModel().getPoolModel().getId());
        MessageFlow inMessageFlow = new MessageFlow();
        inMessageFlow.setId("inMsgFlow" + taskModel.getId());
        inMessageFlow.setMessageRef(operation.getInMessageRef());
        inMessageFlow.setSourceRef(taskQName);
        inMessageFlow.setTargetRef(poolQName);
        context.getCollaboration().getMessageFlows().add(inMessageFlow);

        if (outMessage != null) {
          MessageFlow outMessageFlow = new MessageFlow();
          outMessageFlow.setId("outMsgFlow" + taskModel.getId());
          outMessageFlow.setMessageRef(operation.getOutMessageRef());
          outMessageFlow.setSourceRef(poolQName);
          outMessageFlow.setTargetRef(taskQName);
          context.getCollaboration().getMessageFlows().add(outMessageFlow);
        }

        // Add data input
        IoSpecification ioSpecification = new IoSpecification();
        tTask.setIoSpecification(ioSpecification);

        final String xpathExpression = taskModel.getInputExpression();
        final DataInput dataInput = new DataInput();
        dataInput.setItemSubjectRef(inMessage.getItemRef());
        dataInput.setName("payload");
        dataInput.setId("dataInput" + taskModel.getId());

        ioSpecification.getDataInputs().add(dataInput);
        InputSet inputSet = new InputSet();
        inputSet.getDataInputRefs().add(context.getBpmnObjectFactory().createInputSetDataInputRefs(dataInput));
        ioSpecification.getInputSets().add(inputSet);

        final Assignment assignment = new Assignment();
        final TExpression fromExpression = new TFormalExpression();
        fromExpression.getContent().add(xpathExpression);
        assignment.setFrom(fromExpression);
        final TExpression toExpression = new TFormalExpression();
        toExpression.getContent().add("getDataInput(\"" + dataInput.getName() + "\")");
        assignment.setTo(toExpression);


        final DataInputAssociation dataInputAssociation = new DataInputAssociation();
        // sourceRef elements : Nothing to do with SourceRef in our case (no traduction of data association executable in BPEL)
        //  spec BPMN 2.0 p 253 : In any case, sources are used to define if the data association can be
        // “executed,” if any of the sources is in the state of“unavailable,” then the data association cannot be executed,
        // and the Activity or Event where the data association is defined MUST wait until this condition is met.
        dataInputAssociation.setTargetRef(dataInput);


        dataInputAssociation.getAssignments().add(assignment);
        tTask.getDataInputAssociations().add(dataInputAssociation);

        // Add data output
        if (outMessage != null) {
          final String variableName = taskModel.getOutputVariableName();
          TDataObject variable = null;
          for (final JAXBElement<? extends TFlowElement> flowElement : context.getTProcess().getFlowElements()) {
            if (flowElement.getValue() instanceof TDataObject && flowElement.getValue().getName().equals(variableName)) {
              variable = (TDataObject) flowElement.getValue();
            }
          }
          if (variable == null) {
            variable = new TDataObject();
            variable.setName(variableName);
            variable.setId("var" + taskModel.getId());
            variable.setItemSubjectRef(outMessage.getItemRef());
            context.getTProcess().getFlowElements().add(context.getBpmnObjectFactory().createDataObject(variable));
          }
          // Add data output
          final DataOutput dataOutput = new DataOutput();
          dataOutput.setItemSubjectRef(outMessage.getItemRef());
          dataOutput.setName("payload");
          dataOutput.setId("dataOutput" + taskModel.getId());
          ioSpecification.getDataOutputs().add(dataOutput);
          OutputSet outputSet = new OutputSet();
          outputSet.getDataOutputRefs().add(context.getBpmnObjectFactory().createOutputSetDataOutputRefs(dataOutput));
          ioSpecification.getOutputSets().add(outputSet);

          final DataOutputAssociation dataOutputAssociation = new DataOutputAssociation();
          dataOutputAssociation.getSourceReves().add(context.getBpmnObjectFactory().createTDataAssociationSourceRef(dataOutput));
          dataOutputAssociation.setTargetRef(variable);

          tTask.getDataOutputAssociations().add(dataOutputAssociation);
        }
      }
    } else {
      tTask = new TTask();
      jaxbElement = context.getBpmnObjectFactory().createTask(tTask);
    }
    tTask.setId(taskModel.getId());
    tTask.setName(taskModel.getName());

    context.getModelsMap().put(taskModel, tTask);

    context.getTProcess().getFlowElements().add(jaxbElement);

    // Now deals with positions
    ShapeTransformer.modelToBpmn(taskModel, context);

  }

  /**
   * Create interface and operation for a task.
   *
   * @param taskModel
   * @param context
   * @return
   * @throws TransformationException
   */
  private static Operation modelToBpmnOperation(final TaskModel taskModel, final Designer2BpmnContext context) throws TransformationException {

    final String interfaceName = taskModel.getOperation().getInterfaceModel().getName();
    // Find interface
    TInterface tInterface = null;
    for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
      if (rootElement.getValue() instanceof TInterface && ((TInterface) rootElement.getValue()).getName().equals(interfaceName)) {
        tInterface = (TInterface) rootElement.getValue();
      }
    }
    if (tInterface == null) {
      throw new TransformationException("BPMN conversion error: task " + taskModel.getName() + " use an unknown interface " + interfaceName);
    }

    // Check interface is in collaboration
    if (context.getCollaboration() == null) {
      throw new TransformationException("BPMN conversion error: no collaboration defined for task " + taskModel.getName());
    }
    final TCollaboration collaboration = context.getCollaboration();
    Participant interfaceParticipant = null;
    final QName interfaceQName = new QName(context.getDefinitions().getTargetNamespace(), tInterface.getId());
    for (final Participant participant : collaboration.getParticipants()) {
      if (participant.getInterfaceReves().contains(interfaceQName)) {
        interfaceParticipant = participant;
        break;
      }
    }
    if (interfaceParticipant == null) {
      throw new TransformationException("BPMN conversion error: no participant found for task " + taskModel.getName());
    }
    // Find or create operation
    final String operationName = taskModel.getOperation().getName();
    Operation operation = null;
    for (final Operation interfaceOperation : tInterface.getOperations()) {
      if (interfaceOperation.getName().equals(operationName)) {
        operation = interfaceOperation;
      }
    }
    if (operation == null) {
      throw new TransformationException("BPMN conversion error: no operation " + operationName + " defined for interface " + interfaceName + " (task " + taskModel.getName() + ")");
    }
    return operation;
  }
}
